@@ -1,9 +1,15 @@
-import datetime
 import logging
-import netCDF4 as netcdf
-import numpy as np
-import os
+import datetime
 import pytz
+import os
+
+import numpy as np
+import netCDF4 as netcdf
+
+from anytree import Node
+from anytree.iterators.preorderiter import PreOrderIter
+
+from collections import Counter
 
 logger = logging.getLogger(__name__)
 
@@ -17,6 +23,9 @@ class HiRElPPFile:
         if read_now:
             self.read_file()
 
+    def __repr__(self):
+        return 'HiRElPP File: {}'.format(self.file_path)
+
     def read_file(self):
         """ Read required data from the NetCDF file."""
         with netcdf.Dataset(self.file_path) as f:
@@ -39,6 +48,7 @@ class HiRElPPFile:
                                                                'attenuated_backscatter_detection_wavelength'][:]
         self.attenuated_backscatter_emission_wavelength = f.variables[
                                                               'attenuated_backscatter_emission_wavelength'][:]
+
         try:
             calibration = f.variables['attenuated_backscatter_calibration'][:]
         except:
@@ -62,7 +72,6 @@ class HiRElPPFile:
             self.has_depolarization = False
 
         if self.has_depolarization:
-            self.vldr_channel_names = f.variables['volume_linear_depolarization_ratio_channel_name'][:]
             self.vldr_wavelength = f.variables['volume_linear_depolarization_ratio_wavelength'][:]
             self.vldr_range = f.variables['volume_linear_depolarization_ratio_range'][:]
             self.vldr_scatterers = f.variables['volume_linear_depolarization_ratio_scatterers'][:]
@@ -84,16 +93,26 @@ class HiRElPPFile:
         """
         self.time = f.variables['time'][:]
         self.time_bounds = f.variables['time_bounds'][:]
-        self.z = f.variables['altitude'][0, :]
+
+        try:
+            self.z = f.variables['altitude'][0, :]  # TODO: Assumes that altitude is time-invariant. Fix this.
+        except KeyError:
+            self.z = f.variables['height'][0, :]  # Here the 0 index refers to "angle" dimension.
+
         self.station_altitude = f.variables['station_altitude'][:]
         self.longitude = f.variables['longitude'][:]
         self.latitude = f.variables['latitude'][:]
         self.measurement_ID = f.measurement_ID
         self.hoi_system_ID = f.hoi_system_ID
+
+        self.hoi_configuration_ID = getattr(f, 'hoi_configuration_ID', 0)  # Support for older formats.
+
         self.measurement_start_datetime_str = f.measurement_start_datetime
         self.measurement_stop_datetime_str = f.measurement_stop_datetime
         self.start_time = datetime.datetime.strptime(f.measurement_start_datetime, '%Y-%m-%dT%H:%M:%SZ')
         self.stop_time = datetime.datetime.strptime(f.measurement_stop_datetime, '%Y-%m-%dT%H:%M:%SZ')
+        self.duration = self.stop_time - self.start_time
+
         self.institution = f.institution
         self.system = f.system
         self.station_id = f.station_ID
@@ -112,7 +131,7 @@ class HiRElPPFile:
         self.do_phone = getattr(f, 'Data_Originator_phone', '')
         self.dp_email = getattr(f, 'Data_Originator_email', '')
 
-        self.comments = f.comment
+        self.comments = getattr(f, 'comment', '')  # Support for older formats.
 
         self.source = f.source
         self.title = f.title
@@ -127,6 +146,11 @@ class HiRElPPFile:
 
         self.channel_names = f.variables['attenuated_backscatter_channel_name'][:]
 
+        self.has_depolarization = 'volume_linear_depolarization_ratio' in f.variables.keys()
+
+        if self.has_depolarization:
+            self.vldr_channel_names = f.variables['volume_linear_depolarization_ratio_channel_name'][:]
+
         self.number_of_profiles = len(self.time)
 
         self.dts = np.diff(self.time)  # Should be self.time_bounds[:, 1] - self.time_bounds[:, 0]
@@ -146,6 +170,15 @@ class HiRElPPFile:
         self.t_start = [datetime.datetime.fromtimestamp(t, pytz.UTC) for t in self.time_bounds[:, 0]]
         self.t_stop = [datetime.datetime.fromtimestamp(t, pytz.UTC) for t in self.time_bounds[:, 1]]
 
+    def altitude_size(self, min_altitude, max_altitude):
+        """ Return the length (in bins) between two altitudes. """
+        idx_min = self.idx_at_altitude(min_altitude)
+        idx_max = self.idx_at_altitude(max_altitude)
+        return idx_max - idx_min
+
+    def idx_at_altitude(self, altitude):
+        return np.argmin(np.abs(self.z - altitude))
+
     def read_metadata(self):
         """ Read only metadata from Hirelpp file."""
         with netcdf.Dataset(self.file_path) as f:
@@ -178,16 +211,20 @@ class HiRElPPFile:
     def convert_to_geoms(self, output_dir, location, affiliation=None):
         """ Convert the input file to something looking like GEOMS format.
 
-        The aim is to demonstrate the possiblities and limitation of GEOMS format,
+        The aim is to demonstrate the capabilities and limitation of GEOMS format,
         not to make a full GEOMS-compatible format.
 
         Parameters
         ----------
         output_dir : str
            Directory of output file. Filename is creates automatically.
+        affiliation : str
+           Initials of institute, to be used in filename (e.g. inoe).
         location : str
            Location name, to be used in filename (e.g. bucharest).
         """
+        # TODO: Make affiliation a kwarg affiliation = None
+
         # Affiliation should be something like "inoe", to be used in the filename
         if affiliation is None:
             if self.pi_affiliation_acronym:
@@ -222,12 +259,15 @@ class HiRElPPFile:
 
         self.fill_geoms_file(output_path)
 
-    def fill_geoms_file(self, output_path, min_idx=None, max_idx=None, min_output_idx=None, max_output_idx=None):
+    def fill_geoms_file(self, output_path, min_idx=None, max_idx=None, min_output_idx=None, max_output_idx=None,
+                        altitude_min=None, altitude_max=None, selected_backscatter_channels=None,
+                        selected_depol_channels=None):
         """ Copy hirelpp data to an initialized geoms file. """
 
-        with netcdf.Dataset(output_path, 'a') as geoms_file:
+        altitude_min_idx = self.idx_at_altitude(altitude_min)
+        altitude_max_idx = self.idx_at_altitude(altitude_max)
 
-            # geoms_file = netcdf.Dataset(output_path, 'a')
+        with netcdf.Dataset(output_path, 'a') as geoms_file:
 
             vars = geoms_file.variables  # Shorthand
             altitude = vars['ALTITUDE']
@@ -254,6 +294,12 @@ class HiRElPPFile:
             longitude = vars['LONGITUDE.INSTRUMENT']
             viewing_angle = vars['ANGLE.VIEW_ZENITH']
 
+            if selected_backscatter_channels is None:
+                selected_backscatter_channels = self.channel_names
+
+            backscatter_idxs = np.array(
+                [np.where(self.channel_names == channel_name)[0][0] for channel_name in selected_backscatter_channels])
+
             if self.has_depolarization:
                 depol_channel = vars['VOLUME.LINEAR.DEPOLARIZATION.RATIO_CHANNEL.NAME']
                 vldr = vars['VOLUME.LINEAR.DEPOLARIZATION.RATIO']
@@ -262,51 +308,69 @@ class HiRElPPFile:
                 vldr_scatterers = vars['VOLUME.LINEAR.DEPOLARIZATION.RATIO_SCATTERERS']
                 vldr_wl = vars['VOLUME.LINEAR.DEPOLARIZATION.RATIO_EMISSION.WAVELENGTH']
 
+                if selected_depol_channels is None:
+                    selected_depol_channels = self.vldr_channel_names
+
+                depol_idx = np.array([np.where(self.vldr_channel_names == channel_name)[0][0] for channel_name in
+                                      selected_depol_channels])
+
             # Assign values
-            altitude[:] = self.z
+            altitude[:] = self.z[altitude_min_idx:altitude_max_idx]
             datetime_var[min_output_idx:max_output_idx] = self.epoch_to_mjd2k(np.mean(self.time_bounds, axis=1))[
                                                           min_idx:max_idx]
+
             datetime_start[min_output_idx:max_output_idx] = self.epoch_to_mjd2k(self.time_bounds[:, 0])[min_idx:max_idx]
             datetime_stop[min_output_idx:max_output_idx] = self.epoch_to_mjd2k(self.time_bounds[:, 1])[min_idx:max_idx]
-            backscatter_channel[:] = self.channel_names
+            backscatter_channel[:] = np.array(selected_backscatter_channels)
             latitude[:] = self.latitude
             longitude[:] = self.longitude
             altitude_inst[:] = self.station_altitude
             viewing_angle[:] = self.laser_pointing_angle[0]
+            backscatter_emission_wl[:] = self.attenuated_backscatter_emission_wavelength[backscatter_idxs]
+            backscatter_detection_wl[:] = self.attenuated_backscatter_detection_wavelength[backscatter_idxs]
+            backscatter_range_type[:] = self.attenuated_backscatter_range[backscatter_idxs]
+            backscatter_scatterers[:] = self.attenuated_backscatter_scatterers[backscatter_idxs]
+            backscatter[min_output_idx:max_output_idx, :, :] = np.transpose(self.attenuated_backscatter, (1, 2, 0))[
+                                                               min_idx:max_idx, altitude_min_idx:altitude_max_idx,
+                                                               backscatter_idxs]
+            backscatter_error[min_output_idx:max_output_idx, :, :] = np.transpose(self.attenuated_backscatter_error,
+                                                                                  (1, 2, 0))[min_idx:max_idx,
+                                                                     altitude_min_idx:altitude_max_idx,
+                                                                     backscatter_idxs]
+            backscatter_calibration[min_output_idx:max_output_idx, :] = np.transpose(self.signal_calibration, (1, 0))[
+                                                                        min_idx:max_idx, backscatter_idxs]
+            backscatter_calibration_error[min_output_idx:max_output_idx, :, ] = np.transpose(
+                self.signal_calibration_statistica_error, (1, 0))[min_idx:max_idx, backscatter_idxs]
             backscatter_emission_wl[:] = self.attenuated_backscatter_emission_wavelength
             backscatter_detection_wl[:] = self.attenuated_backscatter_detection_wavelength
             backscatter_range_type[:] = self.attenuated_backscatter_range
             backscatter_scatterers[:] = self.attenuated_backscatter_scatterers
-            backscatter[min_output_idx:max_output_idx, :, :] = np.transpose(self.attenuated_backscatter, (1, 2, 0))[
-                                                               min_idx:max_idx, :, :]
-            backscatter_error[min_output_idx:max_output_idx, :, :] = np.transpose(self.attenuated_backscatter_error,
-                                                                                  (1, 2, 0))[min_idx:max_idx, :, :]
-            backscatter_calibration[min_output_idx:max_output_idx, :] = np.transpose(self.signal_calibration, (1, 0))[
-                                                                        min_idx:max_idx, :]
-            backscatter_calibration_error[min_output_idx:max_output_idx, :, ] = np.transpose(
-                self.signal_calibration_statistica_error, (1, 0))[min_idx:max_idx, :]
 
             if self.has_depolarization:
-                depol_channel[:] = self.vldr_channel_names
-                vldr_wl[:] = self.vldr_wavelength
-                vldr_range_type[:] = self.vldr_range
-                vldr_scatterers[:] = self.vldr_scatterers
-                vldr[min_output_idx:max_output_idx, :, :] = np.transpose(self.vldr, (1, 2, 0))[min_idx:max_idx, :, :]
+                depol_channel[:] = np.array(selected_depol_channels)
+                vldr_wl[:] = self.vldr_wavelength[depol_idx]
+                vldr_range_type[:] = self.vldr_range[depol_idx]
+                vldr_scatterers[:] = self.vldr_scatterers[depol_idx]
+                vldr[min_output_idx:max_output_idx, :, :] = np.transpose(self.vldr, (1, 2, 0))[min_idx:max_idx,
+                                                            altitude_min_idx:altitude_max_idx, depol_idx]
                 vldr_error[min_output_idx:max_output_idx, :, :] = np.transpose(self.vldr_error, (1, 2, 0))[
-                                                                  min_idx:max_idx, :, :]
+                                                                  min_idx:max_idx, altitude_min_idx:altitude_max_idx,
+                                                                  depol_idx]
+                depol_channel[:] = self.vldr_channel_names
 
             atmospheric_background[min_output_idx:max_output_idx, :] = np.transpose(self.background, (1, 0))[
-                                                                       min_idx:max_idx, :]
+                                                                       min_idx:max_idx, backscatter_idxs]
             atmospheric_background_stdev[min_output_idx:max_output_idx, :] = np.transpose(self.background_stdev,
-                                                                                          (1, 0))[min_idx:max_idx, :]
+                                                                                          (1, 0))[min_idx:max_idx,
+                                                                             backscatter_idxs]
             atmospheric_background_sterr[min_output_idx:max_output_idx, :] = np.transpose(self.background_sterr,
-                                                                                          (1, 0))[min_idx:max_idx, :]
+                                                                                          (1, 0))[min_idx:max_idx,
+                                                                             backscatter_idxs]
             atmospheric_background_min[min_output_idx:max_output_idx, :] = np.transpose(self.background_min, (1, 0))[
-                                                                           min_idx:max_idx, :]
+                                                                           min_idx:max_idx, backscatter_idxs]
             atmospheric_background_max[min_output_idx:max_output_idx, :] = np.transpose(self.background_max, (1, 0))[
-                                                                           min_idx:max_idx, :]
+                                                                           min_idx:max_idx, backscatter_idxs]
 
-            # geoms_file.close()
 
     def initialize_geoms_file(self, output_path, time_size, altitude_size, backscatter_size,
                               depolarization_size, start_time, stop_time):
@@ -833,10 +897,10 @@ class HiRElPPFile:
         return mjd
 
 
-class HiRElPPArchive(object):
+class HiRElPPArchive:
 
     def __init__(self, file_list):
-        """ Handle a dataset of HiRElPP files. The aim is to merge serveral hirelpp files
+        """ Handle a dataset of HiRElPP files. The aim is to merge several hirelpp files
         to daily GEOMS files.
 
         Parameters
@@ -844,14 +908,19 @@ class HiRElPPArchive(object):
         file_list : list
            A list of full paths for the HiRElPP files.
         """
-        self.file_list = sorted(file_list)
 
         # Read metadata from all files
-        self.files = []
-        for file_path in self.file_list:
+        files = []
+        for file_path in file_list:
             h = HiRElPPFile(file_path, read_now=False)
             h.read_metadata()
-            self.files.append(h)
+            files.append(h)
+
+        # From https://stackoverflow.com/a/9764364
+        sorted_paths, sorted_files = zip(*sorted(zip(file_list, files), key=lambda x: x[1].start_time))
+
+        self.file_list = list(sorted_paths)
+        self.files = list(sorted_files)
 
         self.start_dates = [h.start_time.date() for h in self.files]
         self.stop_dates = [h.stop_time.date() for h in self.files]
@@ -881,11 +950,16 @@ class HiRElPPArchive(object):
         files = [self.files[c] for c in idxs]
         files.sort(key=lambda x: x.start_time)
 
+        files = self._remove_overlapping(files)
+
         epoch = datetime.date(1970, 1, 1)
         date_timestamp = (date - epoch).total_seconds()
         next_date_timestamp = date_timestamp + 86400
 
         file_metadata = []
+        backscatter_channels = []
+        depol_channels = []
+        altitude_limits = []
 
         last_output_idx = 0
         for h in files:
@@ -910,6 +984,28 @@ class HiRElPPArchive(object):
                             'stop_time': h.t_stop[file_max_idx], }
 
                 file_metadata.append(metadata)
+                backscatter_channels.append(h.channel_names)
+                altitude_limits.append([h.z[0], h.z[-1]])
+
+                if h.has_depolarization:
+                    depol_channels.append(h.vldr_channel_names)
+                else:
+                    depol_channels.append([])
+
+        # Find channels common to all files
+        common_backscatter_channels = set(backscatter_channels[0])
+        for c in backscatter_channels[1:]:
+            common_backscatter_channels.intersection_update(c)
+        common_backscatter_channels = list(common_backscatter_channels)
+
+        common_depol_channels = set(depol_channels[0])
+        for c in depol_channels[1:]:
+            common_depol_channels.intersection_update(c)
+        common_depol_channels = list(common_depol_channels)
+
+        alt_mins, alt_maxs = zip(*altitude_limits)
+        altitude_min = max(alt_mins)
+        altitude_max = min(alt_maxs)
 
         start_time = file_metadata[0]['start_time']
         stop_time = file_metadata[-1]['stop_time']
@@ -927,23 +1023,22 @@ class HiRElPPArchive(object):
         first_file = file_metadata[0]['file']
         first_file.read_file()
 
-        altitude_size = len(first_file.z)
-        backscatter_size = len(first_file.channel_names)
+        altitude_size = first_file.altitude_size(altitude_min, altitude_max)
 
-        if first_file.has_depolarization:
-            depolarization_size = len(first_file.vldr_channel_names)
-        else:
-            depolarization_size = 0
+        backscatter_size = len(common_backscatter_channels)  # len(first_file.channel_names)
+
+        depolarization_size = len(common_depol_channels)
 
         first_file.initialize_geoms_file(output_path, last_output_idx, altitude_size, backscatter_size,
-                                         depolarization_size, start_time_str, stop_time_str)
+                                         depolarization_size, start_time_str, stop_time_str, )
         min_idx = file_metadata[0]['min_idx']
         max_idx = file_metadata[0]['max_idx']
         output_min_idx = file_metadata[0]['output_min_idx']
         output_max_idx = file_metadata[0]['output_max_idx']
 
         logger.debug('Filling with data from the first file.')
-        first_file.fill_geoms_file(output_path, min_idx, max_idx, output_min_idx, output_max_idx)
+        first_file.fill_geoms_file(output_path, min_idx, max_idx, output_min_idx, output_max_idx, altitude_min,
+                                   altitude_max, common_backscatter_channels, common_depol_channels)
 
         for metadata in file_metadata[1:]:
             f = metadata['file']
@@ -956,6 +1051,115 @@ class HiRElPPArchive(object):
             output_max_idx = metadata['output_max_idx']
 
             f.read_file()
-            f.fill_geoms_file(output_path, min_idx, max_idx, output_min_idx, output_max_idx)
+            f.fill_geoms_file(output_path, min_idx, max_idx, output_min_idx, output_max_idx, altitude_min, altitude_max,
+                              common_backscatter_channels, common_depol_channels)
 
         return output_filename
+
+    def _remove_overlapping(self, all_files):
+        """ Remove files with overlapping time periods.
+
+        In case that the files originate from different systems, only files belonging to one system are returned,
+        based on the total measurement duration during the day.
+
+        Parameters
+        ----------
+        all_files : list of HiRElPPFile
+           A sorted list of hirelpp files. """
+
+        system_ids = list(set([f.hoi_system_ID for f in all_files]))
+
+        system_path_parameters = []
+
+        for system_id in system_ids:
+            system_files = [f for f in all_files if f.hoi_system_ID == system_id]
+            system_node = Node(system_id)
+
+            self._fill_tree(system_files, system_node)
+
+            paths, durations, hoi_config_occurrences = self._get_branch_info(system_node)
+
+            if len(paths) == 0:
+                continue
+
+            max_duration_idxs = [n for n, x in enumerate(durations) if x == max(durations)]
+
+            if len(max_duration_idxs) > 1:
+                hoi_occurrences = [hoi_config_occurrences[max_idx] for max_idx in max_duration_idxs]
+                max_occurrence_idx = hoi_occurrences.index(max(hoi_occurrences))
+            else:
+                max_occurrence_idx = 0
+
+            path_idx = max_duration_idxs[max_occurrence_idx]
+
+            chosen_path = paths[path_idx]
+
+            clear_files = [n.name for n in chosen_path[1:]]
+
+            system_path_parameters.append({'system_id': system_id,
+                                           'duration': max(durations),
+                                           'clear_files': clear_files})
+
+        max_path_duration = max([d['duration'] for d in system_path_parameters])
+        max_path_parameters = [d for d in system_path_parameters if d['duration'] == max_path_duration]
+
+        # If more than one system have equal path lengths, return the first one
+        return max_path_parameters[0]['clear_files']
+
+    def _fill_tree(self, hirelpp_files, parent_node):
+
+        overlapping_nodes = []
+        other_files = []
+
+        first_file = hirelpp_files[0]
+        overlapping_nodes.append(Node(first_file, parent=parent_node))
+
+        for f in hirelpp_files[1:]:
+            if self._is_overlapping(first_file, f):
+                overlapping_nodes.append(Node(f, parent=parent_node))
+            else:
+                other_files.append(f)
+
+        if other_files:
+            for node in overlapping_nodes:
+                # Checking for the first node is redundant, but not so expensive anyway
+                non_overlapping_files = self._get_non_overlapping(node, other_files)
+                self._fill_tree(non_overlapping_files, node)
+
+        return parent_node
+
+    def _is_overlapping(self, first_file, second_file):
+        """ Assumes the first_file is starting before the second_file. """
+        return first_file.stop_time > second_file.start_time
+
+    def _get_non_overlapping(self, node, hirelpp_files):
+        """ Return a list of non-overlapping files. """
+        return [f for f in hirelpp_files if not self._is_overlapping(node.name, f)]
+
+    def _get_branch_info(self, base_node):
+        """ Get information about the available paths.
+
+        Exclude paths with variable vertical resolution. """
+        all_paths = [leaf.path for leaf in PreOrderIter(base_node, filter_=lambda node: node.is_leaf)]
+        durations = []
+        hoi_config_occurrences = []
+
+        uniform_paths = []
+
+        for path in all_paths:
+            resolutions = [n.name.dz for n in path[1:]]
+            has_unique_resolution = np.all(np.abs(np.diff(resolutions)) < 1e-3)  # Require mm similarity
+
+            if has_unique_resolution:
+                uniform_paths.append(path)
+
+                durations.append(np.sum([n.name.duration for n in path[1:]]))
+
+                config_list = [n.name.hoi_configuration_ID for n in path[1:]]
+                most_commnon = max(Counter(config_list).values())
+                hoi_config_occurrences.append(most_commnon)
+            else:
+                print("More than one vertical resolutions ({})".format(set(resolutions)))
+
+        return uniform_paths, durations, hoi_config_occurrences
+
